'''
runBinfull.py
A script testing the binfull algorithm on some sample functions.
Copyright (C) 2014 Abram Krislock and Nathan Krislock

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see http://www.gnu.org/licenses/gpl-3.0.txt.
'''
###### IMPORTS ######
from time import time
totalTimer = time()
import utilities.sampleFunctions as funcs
import utilities.minimizers as minis
import binfull

import numpy as np
import bisect
from matplotlib import pyplot as plt
plt.rc('mathtext', fontset='stix')

import sys, getopt, os

setup = {'xlow':0, 'xhigh':200, 'binWidth':5, 'step':0.0001, 'npoints':1000,
         'func':'all', 'dataDir':'dataSet_a', 'seed':'seed_a',
         'nbinfull':100000, 'binfullWidth':0.75, 'smoother':'all'}

###### COMMAND LINE ARGS AND SETUP ######
try:
  opts, args = getopt.getopt(sys.argv[1:], '',
                             [key + '=' for key in setup])
except getopt.GetoptError:
  print 'runBinfull.py may be run with the following options:'
  print '  DATA OPTIONS:'
  print '  --xlow=[a], --xhigh=[b]: Set the x-axis range to [a,b]'
  print '  --binWidth=[#]: Set the width of the bins for the histogram'
  print '  --step=[#]: Set the step-size for fake data generation'
  print '  --npoints=[#]: Set the number of points for fake data generation'
  print '  --func=[#/all]: Choose the function by int from 0 to 4'
  print '  --dataDir=[None/directory]: Data to or from ./savedData/directory'
  print '  --seed=[string]: Set the random seed (uses hash value)'
  print '  BINFULL OPTIONS:'
  print '  --nbinfull=[#]: Set the number of points for binfull'
  print '  --binfullWidth=[#]: Set the width of the bins for binfull'
  print '  --smoother=[#/all]: Choose the smoothing function by int from 0 to 2'
  print '### DEFAULTS:'
  for key, value in setup.iteritems():
    print '  ' + key + '=' + str(value)
  print
  sys.exit(2)
for opt, arg in opts:
  if arg.isdigit():
    setup[opt.strip('-')] = int(arg)
  elif arg.replace('.', '', 1).isdigit():
    setup[opt.strip('-')] = float(arg)
  else:
    setup[opt.strip('-')] = arg

# saving, loading, and plotting settings, creating directories as necessary.
if setup['dataDir']:
  dataDir = 'savedData/' + setup['dataDir'] + '/'
  plotDir = 'resultPlots/' + setup['dataDir'] + '/'
  setupFileName = dataDir.partition(setup['dataDir'])
  setupFileName = setupFileName[0] + 'setup_' + setupFileName[1] + '.txt'

  # create the data directory, if necessary
  if not os.path.isdir(dataDir):
    os.mkdir(dataDir)

  # load the settings, if the settings file exists
  if os.path.exists(setupFileName):
    setupFile = open(setupFileName, 'r')
    line = (setupFile.readline().strip('\n')).partition(' = ')
    while not line == ('', '', ''):
      if line[2].isdigit():
        setup[line[0]] = int(line[2])
      elif line[2].replace('.', '', 1).isdigit():
        setup[line[0]] = float(line[2])
      else:
        setup[line[0]] = line[2]
      line = (setupFile.readline().strip('\n')).partition(' = ')
    setupFile.close()
else:
  plotDir = 'resultPlots/' + S['seed'] + 'N' + str(S['npoints']) + '/'

# make the plotting directory, if necessary
if not os.path.isdir(plotDir):
  os.mkdir(plotDir)

# shorthand for the setup dict
S = setup
# list of functions to test
if S['func'] == 'all':
  funcList = range(len(funcs.allFunctions)) 
else:
  funcList = [S['func']]
# list of smoothing functions to test for each function
if S['smoother'] == 'all':
  smoothList = range(len(binfull.smootherTuple))
else:
  smoothList = [S['smoother']]
# names of the data arrays to be saved or loaded
dataNames = ['x', 'y', 'cdf', 'pdf', 'bgX', 'bgY', 'bgCdf', 'bgPdf']



###### BIG FUNC LOOP ######
for f in funcList:
  # convert func int choice to actual function
  func = funcs.allFunctions[f]

  # filename prefix for this function
  prefix = funcs.nameMap[func]

  # init random number generator with S['seed'] and the function string.
  funcs.randomGenerator.setSeed(S['seed'] + prefix)

  ###### FAKE DATA SETUP ######
  t = time()
  if S['dataDir'] and all(
        [os.path.exists(dataDir + prefix + '_' + name + '.npy')
            for name in dataNames]):
    # fake data exists; load it!
    print '\n ~~~~~~ Now loading function data: '
    print prefix
    print ' with', S['npoints'], 'points ~~~~~~ \n'
    for name in dataNames:
      exec(name + ' = np.load(dataDir + prefix + "_' + name + '.npy")')
  else:
    # generate fake data!
    print '\n ###### Now generating data for function: '
    print prefix
    print ' with ', S['npoints'], 'points ###### \n'
    x, y, cdf, pdf = funcs.functionToData(func, S['npoints'], S['step'])
    bgNpoints = [S['npoints'], cdf[-1]]
    bgX, bgY, bgCdf, bgPdf = funcs.functionToData(funcs.bgMap[func],
                                                  bgNpoints, S['step'])

  setLog = True if funcs.bgMap[func] == funcs.exponentialBG else False
  histogramRange = np.arange(S['xlow'], 0.5*S['binWidth'] + S['xhigh'],
                             S['binWidth'])
  yNormed = y / cdf[-1]
  bgYNormed = bgY / cdf[-1]
  bgNpoints = len(bgPdf)


###### PLOTTING ######
  fig, axes = plt.subplots(figsize=(11,7))
  fig.subplots_adjust(left=0.14, right=0.95, top=0.95, bottom=0.125)
  # save the plot painters as pXxxx
  # the histogram painter is a list of "patches". grab any patch with [2][0]
  pHist = (plt.hist(pdf, bins=histogramRange, log=setLog,
                    normed=True, alpha=0.4, color='#66a61e',
                    label=r'$%i$ Points' % S['npoints'], zorder=0))[2][0]
  pBG, = plt.plot(bgX, bgYNormed, '--', color='#964a01', linewidth=4,
                  label='BG only PDF', zorder=1)
  pTrue, = plt.plot(x, yNormed, '#964a01', linewidth=4,
                    label='True PDF', zorder=2)

  xwidth = S['xhigh'] - S['xlow']
  xdiv = xwidth / 4.
  axes.set_xlim(S['xlow'] - 0.002*xwidth, S['xhigh'] + 0.001*xwidth)
  axes.set_xticks(np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv))
  axes.set_xticklabels([r'$%i$' % int(num) for num in
                        np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv)],
                       fontsize=30)
  axes.set_xticks(np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv/5.),
                  minor=True)

  if setLog:
    axes.set_ylim(bottom=0.97e-4, top=0.101)
    axes.set_yticklabels([r'$10^{%i}$' % int(num) for num in
                          np.arange(-5,1,1)], fontsize=30)
  else:
    # This is a bit of a pain in the ass... but makes a pretty yaxis:
    ymaxx = 1.6 * plt.axis()[3]
    ydiv = 10**(np.log10(ymaxx) // 1)
    if ydiv * 3. > ymaxx:
      ydiv /= 2.
    elif ydiv * 5. <= ymaxx:
      ydiv *= 2.
    axes.set_ylim(bottom=-0.0001, top=ymaxx)
    axes.set_yticks(np.arange(0, ymaxx*1.01, ydiv))
    axes.set_yticklabels([r'$%0.3f$' % num for num
                          in np.arange(0, ymaxx*1.01, ydiv)], fontsize=30)
    axes.set_yticks(np.arange(0, ymaxx*1.01, ydiv/5.), minor=True)

  axes.set_xlabel('Arbitrary Units', labelpad=5, fontsize=28)
  axes.set_ylabel('Probability', labelpad=5, fontsize=28)

  axes.tick_params(length=10, width=2, labelsize=28, zorder=10, pad=10)
  axes.tick_params(which='minor', length=5, width=2, zorder=11)
  axes.minorticks_on()

  plt.legend((pTrue, pBG, pHist),
             (pTrue.get_label(), pBG.get_label(), pHist.get_label()),
             shadow=True, loc='best', prop={'size':26},
             labelspacing=0.25, handlelength=1.2, handletextpad=0.35)

  filename = plotDir + prefix + '_data.pdf'
  plt.savefig(filename)
  plt.close()
  print 'Time after fake data setup: ' + str(time() - t)


###### FAKE DATA SAVING TO FILES ######
  if S['dataDir']:
    t = time()
    for name in dataNames:
      if not os.path.exists(dataDir + prefix + '_' + name + '.npy'):
        eval('np.save(dataDir + prefix + "_' + name + '", ' + name + ')')
    print 'Time after fake data saving: ' + str(time() - t)


###### BINFULL ARRAY SET UP ######
  binfullRange = np.arange(S['xlow'], 0.5*S['binfullWidth'] + S['xhigh'],
                           S['binfullWidth'])
  bgBinfullCenters = np.arange(S['xlow'] + 0.5*S['binfullWidth'], S['xhigh'],
                               S['binfullWidth'])
  # left-most entry in bgX which is >= bgBinfullCenters[i]
  bgXindices = [bisect.bisect_left(bgX, value) for value in bgBinfullCenters]


###### BINFULL PARAMETER FINDING FOR EACH SMOOTHER ######
  for s in smoothList:
    # convert smoother int choice to actual smoother
    smoother = binfull.smootherTuple[s]
    smootherInstance = smoother(1.)

    # filename prefix for this smoother and function
    binfullPrefix = prefix + '_binfull' + str(smootherInstance)

###### BINFULL GENERATION ######
    t = time()
    if all([S['dataDir'], binfullPrefix + 'Sigma' in setup,
            os.path.exists(dataDir + binfullPrefix + 'Edges.npy'),
            os.path.exists(dataDir + binfullPrefix + 'Histo.npy')]):
      # binfull data exists; load it!
      print '\n ~~~~~~ Now loading binfull data from: '
      print binfullPrefix + 'Smoother'
      print ' ~~~~~~ \n'
      binfullPdf = binfull.binfullFactory(
                   np.load(dataDir + binfullPrefix + 'Edges.npy'),
                   np.load(dataDir + binfullPrefix + 'Histo.npy'))
      sigma = setup[binfullPrefix + 'Sigma']
    else:
      # binfull gets its own random number generator and seed
      binfull.randomGenerator.setSeed(S['seed'] + str(smootherInstance))

      # Use a chisq fit to find the best smoothing parameter
      def minimizeMe(sigma):
        ti = time()
        localSmoother = smoother(sigma)
        bgBinfullPdf = binfull.binfullHistogram(binfullRange, bgPdf, S['nbinfull'],
                                                localSmoother)
        # Normalize the true function -up- to the huge size of the binfull cdf
        # NOTE: normalizing up to avoid possible machine epsilon problems
        normalization = bgBinfullPdf.integral() / bgCdf[-1]
        bgYscaled = bgY * normalization

        # Calculate the chisq for this to match the true function
        chisq = 0
        for i, value in enumerate(bgBinfullPdf.histo):
          chisq += (bgYscaled[bgXindices[i]] - value) ** 2

        print 'This evaluation:'
        print '  ' + str(localSmoother), 
        print 'smoother with sigma:', sigma
        print '  chisq:', chisq
        print 'Time after mimimizeMe evaluation: ' + str(time() - ti)
        del localSmoother, bgBinfullPdf
        return chisq

      print '\n ###### Now running binfull algorithm with:'
      print binfullPrefix + 'Smoother'
      print ' ###### \n'

      sigma = minis.goldenSectionSearch(minimizeMe,
                                        smootherInstance['range'], 0.001)
      print
      print 'Minimization complete!! Best smoothing sigma:', sigma

      setup[binfullPrefix + 'Sigma'] = sigma
      smootherInstance = smoother(sigma)
      binfullPdf = binfull.binfullHistogram(binfullRange, pdf, S['nbinfull'],
                                            smootherInstance)
    print 'Time after binfull setup: ' + str(time() - t)
    print


    t = time()
###### PLOTTING ######
    fig, axes = plt.subplots(figsize=(11,7))
    fig.subplots_adjust(left=0.14, right=0.95, top=0.95, bottom=0.125)
    # save the plot painters as pXxxx, or use proxy painters:
    # the histogram painter is a list of "patches". grab any patch with [2][0]
    pHist = (plt.hist(pdf, bins=histogramRange, log=setLog,
                      normed=True, alpha=0.4, color='#66a61e',
                      label=r'$%i$ Points' % S['npoints'], zorder=0))[2][0]
    pTrue, = plt.plot(x, yNormed, '#964a01', linewidth=4,
                      label='True PDF', zorder=3)

    ex = int(np.log10(sigma) // 1)
    if ex:
      sigma *= 10**-ex
      sigma = r'$\sigma = %0.3f \times 10^{%i}$' % (sigma, ex)
    else:
      sigma = r'$\sigma = %0.3f$' % sigma
    proxySigma, = plt.plot(x[:1], yNormed[:1], 'w', linewidth=0, zorder=1,
                           label=sigma)
    proxyBinfull = binfullPdf.fillPlot(color='#a41666', alpha=0.55, zorder=2,
                      label=str(smootherInstance)+' Binfull')

    axes.set_xlim(S['xlow'] - 0.002*xwidth, S['xhigh'] + 0.001*xwidth)
    axes.set_xticks(np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv))
    axes.set_xticklabels([r'$%i$' % int(num) for num in
                          np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv)],
                         fontsize=30)
    axes.set_xticks(np.arange(S['xlow'], S['xhigh'] + 0.001*xwidth, xdiv/5.),
                    minor=True)

    if setLog:
      axes.set_ylim(bottom=0.97e-4, top=0.101)
      axes.set_yticklabels([r'$10^{%i}$' % int(num) for num in
                            np.arange(-5,1,1)], fontsize=30)
    else:
      axes.set_ylim(bottom=-0.0001, top=ymaxx)
      axes.set_yticks(np.arange(0, ymaxx*1.01, ydiv))
      axes.set_yticklabels([r'$%0.3f$' % num for num
                            in np.arange(0, ymaxx*1.01, ydiv)], fontsize=30)
      axes.set_yticks(np.arange(0, ymaxx*1.01, ydiv/5.), minor=True)

    axes.set_xlabel('Arbitrary Units', labelpad=5, fontsize=28)
    axes.set_ylabel('Probability', labelpad=5, fontsize=28)

    axes.tick_params(length=10, width=2, labelsize=28, zorder=10, pad=10)
    axes.tick_params(which='minor', length=5, width=2, zorder=11)
    axes.minorticks_on()

    plt.legend((pTrue, pHist, proxyBinfull, proxySigma),
               (pTrue.get_label(), pHist.get_label(),
                proxyBinfull.get_label(), proxySigma.get_label()),
               shadow=True, loc='best', prop={'size':26},
               labelspacing=0.25, handlelength=1.2, handletextpad=0.35)

    filename = plotDir + binfullPrefix + '.pdf'
    plt.savefig(filename)
    plt.close()

    print 'Time after plotting: ' + str(time() - t)
    print


###### BINFULL SAVING TO A FILE ######
    if S['dataDir']:
      t = time()
      if not os.path.exists(dataDir + binfullPrefix + 'Edges.npy'):
        np.save(dataDir + binfullPrefix + 'Edges', binfullPdf.binEdges)
      if not os.path.exists(dataDir + binfullPrefix + 'Histo.npy'):
        np.save(dataDir + binfullPrefix + 'Histo', binfullPdf.histo)
      print 'Time after binfull saving: ' + str(time() - t)
      print

###### FINALIZATIONS ######
# save the settings, and the sigma results:
if S['dataDir']:
  setupFile = open(setupFileName, 'w')
  for key, value in setup.iteritems():
    setupFile.write(key + ' = ' + str(value) + '\n')
  setupFile.close()

print 'Time after everything: ' + str(time() - totalTimer)
